/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.trigger.impl;

import java.io.IOException;
import java.io.PrintStream;
import java.util.Calendar;
import java.text.SimpleDateFormat;
import java.util.Hashtable;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.rmi.RemoteException;

import javax.xml.namespace.QName;

import org.apache.axis.encoding.AnyContentType;
import org.apache.axis.message.Text;
import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.utils.XMLUtils;
import org.apache.axis.components.uuid.UUIDGen;
import org.apache.axis.components.uuid.UUIDGenFactory;

import org.apache.xpath.XPathAPI;
import org.apache.xpath.objects.XBoolean;
import org.apache.xpath.objects.XNodeSet;
import org.apache.xpath.objects.XNumber;
import org.apache.xpath.objects.XObject;
import org.apache.xpath.objects.XString;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.mds.aggregator.impl.AggregatorServiceGroupResource;
import org.globus.mds.aggregator.impl.AggregatorServiceGroupEntryResource;
import org.globus.mds.aggregator.impl.ServiceGroupEntryAggregatorSink;
import org.globus.mds.aggregator.impl.QueryAggregatorSource;
import org.globus.mds.aggregator.impl.AggregatorSource;
import org.globus.mds.aggregator.impl.AggregatorSink;

import org.globus.mds.aggregator.types.AggregatorContent;
import org.globus.mds.aggregator.types.AggregatorConfig;
import org.globus.mds.aggregator.types.AggregatorData;
import org.globus.mds.aggregator.types.PairedKeyType;

import org.globus.mds.trigger.types.ActionResult;
import org.globus.mds.trigger.types.TriggerRuleType;
import org.globus.mds.trigger.types.TriggerStatusType;
import org.globus.mds.trigger.types.TriggerStatusArray;
import org.globus.mds.trigger.types.TriggerInvalidityTime;

import org.globus.mds.usefulrp.AccumulatorResourceProperty;

import org.globus.wsrf.config.ContainerConfig;

import org.globus.wsrf.Constants;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.TopicList;
import org.globus.wsrf.TopicListAccessor;

import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;

import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleTopicList;
import org.globus.wsrf.impl.SimpleResourcePropertyMetaData;

import org.globus.wsrf.impl.servicegroup.ServiceGroupConstants;
import org.globus.wsrf.impl.servicegroup.EntryResourcePropertyTopic;

import org.globus.wsrf.utils.AnyHelper;
import org.globus.wsrf.utils.XmlUtils;
import org.globus.wsrf.utils.Resources;

import org.globus.util.I18n;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.w3c.dom.CharacterData;

import org.xml.sax.SAXException;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.axis.MessageContext;
/**
 * Implementation of a Trigger ServiceGroup resource. Each resource is also
 * an aggregator sink, in order to match trigger rules against received 
 * data.
 */
public class TriggerResource extends AggregatorServiceGroupResource
                             implements AggregatorSink, TopicListAccessor
{
    private static Log logger =
            LogFactory.getLog(TriggerResource.class.getName());

    private ResourceKey resourceKey = null;

    public static final String TRIGGER_NS = 
        "http://mds.globus.org/2004/08/trigger";
    
    public static final QName RP_SET = 
        new QName(TRIGGER_NS, "TriggerRP");
    
    // TODO: need to give the trigger its own RP doc to include this
    public static final QName FIRE_COUNT_RP = 
        new QName(TRIGGER_NS, "ActionsFiredCount");

    private static final UUIDGen uuidGen = UUIDGenFactory.getUUIDGen();
    
    private static I18n i18n = I18n.getI18n(Resources.class.getName());
 
    protected Calendar terminationTime;
    
    private Hashtable executionTable;
    private ResourcePropertySet propSet;
    private TopicList topicList;
    private String [] strings = null;
    private int fooctr=0;
    private TriggerConfiguration triggerConfig = null;
    private String baseDir;

    
    AccumulatorResourceProperty fireCountRP;
 
    public TriggerResource() 
    {
        super.init(TriggerResource.RP_SET);
        
        this.baseDir = getBaseDirectory();
        this.propSet = this.getResourcePropertySet();
        this.topicList = new SimpleTopicList(this);
        ResourceProperty prop = null;
        
        EntryResourcePropertyTopic rpTopic = 
            new EntryResourcePropertyTopic(
                this.propSet.get(ServiceGroupConstants.ENTRY));
        
        this.propSet.add(rpTopic);
        this.topicList.addTopic(rpTopic);
        
        try {
            prop = 
                new ReflectionResourceProperty(
                    SimpleResourcePropertyMetaData.TERMINATION_TIME,
                    this);
            this.propSet.add(prop);
            prop = 
                new ReflectionResourceProperty(
                    SimpleResourcePropertyMetaData.CURRENT_TIME, 
                    this);
            this.propSet.add(prop);


            fireCountRP=new AccumulatorResourceProperty(FIRE_COUNT_RP);

            this.propSet.add(prop);

            // initialize aggregator 
            this.loadCompatibleSources(this);
            
            this.executionTable = 
                this.getTriggerConfiguration().getExecutableMappingTable(); 
            if (logger.isDebugEnabled()) {
                logger.debug("Trigger resource contains the following executable mapping table: " 
                             + this.executionTable.toString());
            }                       
        } catch (Exception e) {
            logger.error("Error during TriggerResource creation: ",e);
        }       
    }

    public TriggerConfiguration getTriggerConfiguration() 
    {
        if (this.triggerConfig != null) {
            return this.triggerConfig;
        }
        
        MessageContext ctx = MessageContext.getCurrentContext();

        String servicePath = ctx.getTargetService();

        String jndiName = Constants.JNDI_SERVICES_BASE_NAME + servicePath +
               "/triggerConfiguration";
        if (logger.isDebugEnabled()) {
            logger.debug("Will read configuration from jndi name: "+jndiName);
        }
        try {
            Context initialContext = new InitialContext();
            return this.triggerConfig =
                (TriggerConfiguration) initialContext.lookup(jndiName);

        } catch (NamingException e) {
            logger.error("when performing JNDI lookup for "+jndiName+": "+e );
        } catch (Exception e) {
            logger.error("when performing JNDI lookup for "+jndiName+": "+e );
        }
        return this.triggerConfig;
    }

    /** Used to inform the resource of its key. */
    public void setResourceKey(ResourceKey k) 
        throws Exception
    {
        if(resourceKey==null) {
            resourceKey = k;
        } else {
            throw new Exception("Resource key can only be set once.");
        }
    }

    public ResourceKey getResourceKey() {
        return resourceKey;
    }


    public void setTerminationTime(Calendar time) {
        if (logger.isDebugEnabled()) {
            logger.debug("Set Termination time called: " + time.getTime());
        }
        this.terminationTime = time;
    }
    
    public Calendar getTerminationTime() {
        return this.terminationTime;
    }

    public Calendar getCurrentTime() {
        return Calendar.getInstance();
    }

    public TopicList getTopicList() {
        return this.topicList;
    }



    // AggregatorSink methods

    /** Takes delivery of a message from an AggregatorSource and
      * applies the appropriate rules to it.  
      *
      * @param messageObj the message against which any registered triggerRules
      *  will be applied against
      * 
      * @param entry this AggregatorServiceGroupEntryResource should 
      *  contain appropriate trigger service rules in the AggregatorConfig 
      *  block of the ServiceGroupEntry Content member.
      *
      * TODO: should there be any queuing/locking to stop multiple
      * triggers (either for same rule or for all rules) from firing
      * simultaneously?
      *
      * TODO: add a thread pool for executing trigger actions,
      * break up the monolith
      */
    public void deliver(AnyContentType messageObj, 
                        AggregatorServiceGroupEntryResource entry)
    {
        try 
        {
        if (logger.isDebugEnabled()) {
            logger.debug("In trigger service deliver, with message "+
            AnyHelper.toSingleString(messageObj));
        }

        // fail fast if nothing to do
        MessageElement[] messages = messageObj.get_any();
        if (messages == null) {
            if (logger.isDebugEnabled()) {
                logger.debug("Trigger resource will not process null message(s), no action(s) will be taken");
            }
            return;
        }

        // form an input document from the inbound message elements
        // the triggerRule XPath expressions will be applied against this doc
        Document inputDoc = null;
        try 
        {         
            inputDoc = 
                this.anyToNewDocument(messageObj,
                                      TRIGGER_NS,
                                      "trigger:inputMessage");
        } 
        catch (Exception e) 
        {
            logger.error("Error assembling input document. No action will be taken"); 
            return;
        }        
        if (!inputDoc.getDocumentElement().hasChildNodes()) 
        {
            logger.warn("Input message contains no data. No action will be taken");
            return;
        }                        
                
        // get RP set from entry
        ResourcePropertySet rpSet = entry.getResourcePropertySet();

        // get content RP from entry RP set
        ResourceProperty contentRP = rpSet.get(ServiceGroupConstants.CONTENT);
        
        // extract the content from the service group entry       
        AggregatorContent content = entry.getContent();
        AggregatorConfig aggregatorConfig = content.getAggregatorConfig();
        AggregatorData aggregatorData = content.getAggregatorData();
         
        // the aggregatorData array in a trigger can only ever have one value, 
        // that being the global TriggerStatusArray of all trigger status objects
        TriggerStatusType[] statusArray;
        TriggerStatusArray triggerStatusArray;
        
        try {
            triggerStatusArray = (TriggerStatusArray)
                ObjectDeserializer.toObject(aggregatorData.get_any()[0],
                                            TriggerStatusArray.class);
        }
        catch(Exception e) 
        {
            // if this fails (for example, initially the aggregated data
            // may not be a TriggerStatusArray, or may be null), start off 
            // with a fresh new triggerStatusArray. 
            //
            // if the initial aggregator data is anything other than null, 
            // log a warning.
            triggerStatusArray = new TriggerStatusArray();
            
            if (e instanceof NullPointerException) {
                if (logger.isDebugEnabled()) {
                    logger.debug("Unable to locate existing TriggerStatusArray, creating new trigger status object array");
                }
            } else {
                logger.warn("Unexpected exception while deserializing AggregatorData into internal TriggerStatusArray" + e);
            }
        }    
        
        // populate with the current triggerStatusArray
        Hashtable triggerStatusTable = new Hashtable();
        statusArray = triggerStatusArray.getTriggerStatus();
        if (statusArray != null) {
            for (int i=0;i<statusArray.length;i++) {
                triggerStatusTable.put(
                    statusArray[i].getTriggerID(), statusArray[i]);
            }
        }
        
        //
        // try to load any trigger rules from the AggregatorConfig
        // multiple trigger rules per entry configuration are supported
        // each trigger rule submitted will be tagged with a UUID and stored
        // with a list of matching TriggerStatusTypes
        //
        // any other AggregatorConfig elements encountered in the message 
        // will be stored back as unmodified service group content
        //
        MessageElement[] any = aggregatorConfig.get_any();
        ArrayList configResults = new ArrayList();
        
        if (logger.isDebugEnabled()) {
            logger.debug("Checking config for trigger rule(s):  "+
                AnyHelper.toSingleString(any));
        }
        
        TriggerRuleType triggerRule = null; 
        TriggerStatusType triggerStatus = null;
        int triggerRules = 0;
        
        // go over each config entry to see if it is a trigger rule
        // if it is a trigger rule, try to process the trigger rule
        // and its action script, and then continue on to the next trigger rule
        for(int i=0;i<any.length;i++) 
        {
            if (ObjectDeserializer.contains(
                        any[i],
                        TriggerRuleType.class)) 
            {
    
                try {
                    triggerRule = (TriggerRuleType)
                            ObjectDeserializer.toObject(
                                any[i],
                                TriggerRuleType.class
                            );                 

                } catch(DeserializationException de) { 
                    logger.warn("Could not deserialize trigger rule type: "+de);
                    continue;
                }
                
                // locate the internal triggerStatus associated with this
                // triggerRule using the UUID of the triggerRule, create a
                // new triggerRule and triggerStatus if necessary
                String triggerID = triggerRule.getTriggerID();
                if (triggerID == null) 
                {
                    // if there is no existing triggerID, then it is a newly 
                    // registered entry where the trigger status has not been 
                    // initialized, so create the trigger UUID and a new 
                    // TriggerStatus object for the TriggerRuleType                    
                    triggerID = uuidGen.nextUUID();
                    triggerRule.setTriggerID(triggerID);
                    triggerStatus = new TriggerStatusType();
                    triggerStatus.setTriggerID(triggerID);
                    triggerStatus.setMatchingRule(triggerRule.getMatchingRule());
                    triggerStatusTable.put(triggerID, triggerStatus);                    
                }
                else
                {
                    // if we cant find an existing status for this triggerRule
                    // then that is fatal error at this point
                    // for now, log a warning and try to continue...                    
                    triggerStatus = 
                        (TriggerStatusType) triggerStatusTable.get(triggerID);

                    if (triggerStatus == null) {
                        logger.warn("Internal error: Unable to locate trigger status for a registered trigger rule. The trigger rule " + 
                                    triggerID + " is not valid and will not be processed.");
                        continue;
                    }
                }
                
                // store the modified trigger rule
                configResults.add(
                    (MessageElement)ObjectSerializer.toSOAPElement(
                        triggerRule,
                        triggerRule.getTypeDesc().getXmlType())
                );                
                triggerRules++;                
            }
            else 
            {
                configResults.add(any[i]);
                continue;
            }
                        
            // extract xpath expression
            String xpath = triggerRule.getMatchingRule();
            if (logger.isDebugEnabled()) {
                logger.debug(
                    "Processing trigger rule, XPath expression is: " + xpath);
            }

            boolean evaluateBoolean = false;
            Boolean bool = triggerRule.getEvaluateBoolean();
            if (bool != null) {
                evaluateBoolean = bool.booleanValue();
            }            
            
            // match expression
            // if match  (TODO we also need handling of when a condition CEASES to be true)
            boolean fire = false;
            triggerStatus.setRuleLastCheckedAt(Calendar.getInstance());

            Document outputDoc = null;
            try {
                outputDoc = 
                    this.evaluateXPath(inputDoc,
                                       xpath,
                                       triggerRule.getNamespaceMappings(), 
                                       evaluateBoolean);  
            }
            catch(Exception e) 
            {
                String error = "Trigger was not eligible to fire due to exception while evaluating XPath rule: " + e;
                logger.warn( error);  // maybe not log a warning here?
                triggerStatus.setStatusInfo(error);
                continue;
            } 

            if(outputDoc!=null) 
            {
                // we fire the trigger if any results were returned.                          
                if (outputDoc.getDocumentElement().hasChildNodes()) 
                {                
                    if (logger.isDebugEnabled()) {                
                        logger.debug("Trigger scheduled to fire! XPath query result was: " + 
                                     XmlUtils.toString(outputDoc));
                    }
                    fire=true;
                }
                else
                {
                    // the XPath returned no matching results, set the status 
                    // desc to indicate this
                    String msg = "XPath query returned no results (or false) - trigger was not eligible to fire";
                    triggerStatus.setStatusInfo(msg);
                    if (logger.isDebugEnabled()) { 
                        logger.debug(msg);
                    }
                }
            }
            Calendar now = Calendar.getInstance();

            if (!fire)
            {
                triggerStatus.setConditionTrueSince(null);
            }

            if (fire && triggerStatus.getConditionTrueSince() == null) {
                triggerStatus.setConditionTrueSince(now);
            } 

            // perform rate limiting here. This if() will set fire back to
            // false if rate limiting applies.
            if (fire) 
            {
                //   if we are set to fire, then rate limit if:
                //              the conditionTrue has not been true 
                //              for long enough, 
                //              (now - conditionTrueSince) < minimumMatchTime

                if (triggerStatus.getConditionTrueSince()!=null &&
                    triggerRule.getMinimumMatchTime()!=null )  
                {
                    long conditionTrueDuration = 
                        now.getTime().getTime() - 
                            triggerStatus.getConditionTrueSince().getTime().getTime();

                    long minimumMatchTime = triggerRule.getMinimumMatchTime().intValue()*1000;

                    if (conditionTrueDuration < minimumMatchTime) 
                    {
                        fire=false;
                        String msg = 
                            "Inhibited trigger action because condition not true for sufficient time: " + 
                            "the condition has been true for " + 
                            Long.toString(conditionTrueDuration) + 
                            " milliseconds, while the minimum match time is set to " + 
                            Long.toString(minimumMatchTime) + " milliseconds.";
                        
                        triggerStatus.setStatusInfo(msg);
                        if (logger.isDebugEnabled()) {                
                            logger.debug(msg);
                        }
                    }
                    else // mark the last time the condition was true
                    {
                        Calendar last = Calendar.getInstance();
                        last.setTimeInMillis(
                            now.getTime().getTime() - minimumMatchTime);
                        triggerStatus.setConditionTrueSince(last);
                    }
                }
                //       or
                //              we have fired recently
                //              (now - actionFiredAt ) < minimumFiringInterval
                if( triggerStatus.getActionFiredAt()!= null &&
                    triggerRule.getMinimumFiringInterval()!= null)
                {
                    long actionFiredInterval = 
                        now.getTime().getTime() - 
                            triggerStatus.getActionFiredAt().getTime().getTime();

                    long minimumFiringInterval = 
                        triggerRule.getMinimumFiringInterval().intValue()*1000;

                    if (actionFiredInterval < minimumFiringInterval)
                    {
                        fire=false;
                        
                        String msg = 
                            "Inhibited trigger action because " + 
                            "minimumFiringInterval not met: the action was triggered " + 
                            Long.toString(actionFiredInterval) + 
                            " milliseconds ago, while the minimum firing interval is set to " + 
                            Long.toString(minimumFiringInterval) + " milliseconds.";
                        
                        triggerStatus.setStatusInfo(msg);
                        
                        if (logger.isDebugEnabled()) {
                            logger.debug(msg);                        
                        }
                    }
                }
            }           
            
            // if we are scheduled to fire, but the current time falls within
            // any registered invalidity time for this trigger, do NOT fire
            if (fire) 
            {
                TriggerInvalidityTime[] invalidity = 
                    triggerRule.getInvalidityTime();
                
                if ((invalidity != null) && (invalidity.length > 0)) 
                {
                    Calendar startTime = null;
                    Calendar endTime = null;
                    
                    for (int n=0; n < invalidity.length; n++) 
                    {
                        startTime = invalidity[n].getStartTime();
                        endTime = invalidity[n].getEndTime();
                        
                        if ((now.after(startTime)) && (now.before(endTime))) 
                        {
                            SimpleDateFormat sdf = 
                                new SimpleDateFormat(
                                    "EEE d MMM yyyy HH:mm:ss zZ");
                            
                            triggerStatus.setStatusInfo(
                                "The trigger action was inhibited because the execution time fell within the following invalidity time range: " + 
                                 "[" + sdf.format(startTime.getTime()) + 
                                 "] to [" + sdf.format(endTime.getTime()) +"]");

                            fire=false;
                            break;
                        }
                    }
                }
            }
            
            // parse action script input options
            boolean disableUnmodifiedActionScriptInput = false;
            boolean enableFilteredActionScriptInput = false;
            
            bool = triggerRule.getDisableUnmodifiedActionScriptInput();
            if (bool != null) {
                disableUnmodifiedActionScriptInput = bool.booleanValue();
            }
            bool = triggerRule.getEnableFilteredActionScriptInput();
            if (bool != null) {
                enableFilteredActionScriptInput = bool.booleanValue();
            }
            
            // execute the trigger action process
            if(fire) 
            {                
                try 
                {
                    Element actionResultElement = 
                        runAction(triggerRule, 
                            disableUnmodifiedActionScriptInput ? null : inputDoc,
                            enableFilteredActionScriptInput ? outputDoc : null,
                            entry,
                            aggregatorConfig);

                    triggerStatus.setActionFiredAt(Calendar.getInstance());
                    fireCountRP.accumulate(1);

                    ActionResult actionResult = new ActionResult();

                    MessageElement[] m = new MessageElement[] {
                        new MessageElement(actionResultElement)
                    };

                    actionResult.set_any(m);

                    triggerStatus.setActionOutput(actionResult);  

                    String msg = "The trigger action was executed successfully";
                    triggerStatus.setStatusInfo(msg);                    
                }
                catch (Exception e) 
                {
                    String error = "Exception while attempting to execute the trigger action: " + e;
                    logger.warn(error);
                    triggerStatus.setStatusInfo(error);      
                }
                
            }            
        } // end if multiple TriggerRuleType loop

        if (triggerRules == 0) {
            logger.info("A message was delivered, but no valid trigger rules were present in the configuration. No actions were taken.");
            return;
        }  

        // update content            
        try 
        {
            // first, store the config which links TriggerRuleTypes 
            // to TriggerStatusTypes            
            aggregatorConfig.set_any( 
                (MessageElement[]) 
                    configResults.toArray(new MessageElement[] {} )
            );
        }
        catch (Exception e) 
        {
            logger.error("Exception while updating configuration into service group content", e);
        }
             
        try
        {
            // Now store all of the trigger status state               
            triggerStatusArray.setTriggerStatus(
                (TriggerStatusType[]) 
                    triggerStatusTable.values().toArray(
                        new TriggerStatusType[] {} )
            );
            
            aggregatorData.set_any( 
                new MessageElement[] {
                    (MessageElement)ObjectSerializer.toSOAPElement(
                        triggerStatusArray, 
                        triggerStatusArray.getTypeDesc().getXmlType()) 
                } 
            );
            
            entry.setContent(content);            
        } 
        catch(Exception e) 
        {
            logger.error("Exception while updating trigger action result output into service group content", e);
        }         
        
        if (logger.isDebugEnabled()) {
            logger.debug("End of trigger resource deliver");
        }
        
        }
        catch (Exception e)
        {
            logger.error("Unhandled exception in trigger resource deliver method: " + e);
        }
    }

    public Document evaluateXPath(Document inputDoc,
                                  String xPathExpression,
                                  String[] namespaces,
                                  boolean evaluateBoolean) 
        throws Exception
    {
        Document outputDoc = XmlUtils.newDocument();
        Node outputRoot =
            outputDoc.appendChild(
                outputDoc.createElementNS(TRIGGER_NS, "trigger:xpathResult"));
        
        Document namespaceDoc = XmlUtils.newDocument();
        namespaceDoc.appendChild(
            namespaceDoc.createElementNS(TRIGGER_NS, "trigger:namespaceMap"));
        
        if (namespaces != null) 
        {
            if (logger.isDebugEnabled()) {
                logger.debug("Namespace mappings detected...");
            }            
            this.setNamespaces(namespaceDoc.getDocumentElement(), namespaces);
        }        
        
        if (logger.isDebugEnabled()) {
            logger.debug("Applying xpath " + 
                         xPathExpression +
                         " to " + 
                         XmlUtils.toString(inputDoc));
        }
        
        XObject result = null;
        try {
            result = XPathAPI.eval(inputDoc.getDocumentElement(),
                                   xPathExpression,
                                   namespaceDoc.getDocumentElement());
        }
        catch (Exception e) {
            throw new Exception("Exception during trigger matchingRule XPath statement evaluation: " + e); 
        }

        if (result instanceof XBoolean)
        {
            boolean outputBoolean = true;
            if (evaluateBoolean) {
                if (!((XBoolean)result).bool()) {
                    outputBoolean = false;
                }
            } 
            if (outputBoolean) {
                outputRoot.appendChild(
                    outputDoc.importNode(new Text( result.str()),true));
            }

        }
        else if (result instanceof XNumber ||
                 result instanceof XString) 
        {
            MessageElement element = 
                new MessageElement(new Text( result.str() ));
            outputRoot.appendChild(outputDoc.importNode(element,true));
        }
        else if (result instanceof XNodeSet) 
        {
            XNodeSet set = (XNodeSet)result;
            NodeList list = set.nodelist();
            for (int i=0;i<list.getLength();i++) {
                Node node = list.item(i);
                if (node instanceof Document) {
                     outputRoot.appendChild(
                        outputDoc.importNode(
                            new MessageElement(
                                ((Document)node).getDocumentElement()), true));
                } else if (node instanceof Element) {
                     outputRoot.appendChild(
                        outputDoc.importNode(
                            new MessageElement((Element)node), true));
                } else if (node instanceof Attr) {
                    outputRoot.appendChild(
                        outputDoc.importNode(
                            new MessageElement(
                                new Text(
                                    ((Attr)node).getNodeValue() + " ")), true));                        
                } else if (node instanceof CharacterData) {
                     outputRoot.appendChild(
                        outputDoc.importNode(
                            new MessageElement(
                                new Text(
                                    ((CharacterData)node).getData())), true ));
                } else {
                    throw new Exception(i18n.getMessage(
                        "unsupportedXpathReturn", node.getClass().getName()));
                }
            }
        }
        else 
        {
            throw new Exception(i18n.getMessage(
                "unsupportedXpathReturn", result));
        }

        return outputDoc;    
    }  

    private void setNamespaces(Element node, String[] namespaces) 
    {
        if ((namespaces != null) && (namespaces.length > 0)) 
        {
            for (int i=0;i<namespaces.length;i++) 
            {
                if (logger.isDebugEnabled()) {
                    logger.debug(
                        "Adding user-specified namespace to search context: " + 
                        namespaces[i]);
                }
                StringTokenizer st = new StringTokenizer(namespaces[i], "=");
                if ((st.countTokens() % 2) != 0) {
                    logger.warn("Error parsing user-supplied namespace mapping: invalid token in namespace prefix and uri - check string format.");
                    continue;
                }
                while (st.hasMoreTokens()) {
                    node.setAttributeNS("http://www.w3.org/2000/xmlns/",
                                        st.nextToken(),
                                        st.nextToken());
                }
            }
        }
    }         
    
    /** 
        Performs the appropriate trigger action. 
    */
    Element runAction(TriggerRuleType rule,
                      Document inputDoc,
                      Document outputDoc,
                      AggregatorServiceGroupEntryResource entry, 
                      AggregatorConfig config)
        throws Exception
    {        
        if (logger.isDebugEnabled()) {
            logger.debug("Performing trigger action.");
        }

        // $GLOBUS_LOCATION /libexec/trigger/ <actionname>
        String commandline="";

        commandline += baseDir;  //GLOBUS LOCATION

        commandline += "/libexec/trigger/";

        String triggerName = rule.getActionScript();
        String triggerFile = (String)this.executionTable.get(triggerName);              

        if (triggerFile == null) {
            throw new Exception(
                "Could not locate an executable file name mapped to the specified trigger name (check service jndi-config.xml): " 
                + triggerName);
        }
        
        commandline += triggerFile; 
        if (logger.isDebugEnabled()) {
            logger.debug("Will execute: "+commandline);
        }

        // create the action script input document
        // TODO make this into proper XSD   
        StringBuffer actionInput = new StringBuffer();
        actionInput.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
        actionInput.append("<trigger:actionInput xmlns:trigger=\"http://mds.globus.org/2004/08/trigger\">");
        
        if (inputDoc != null) {
            actionInput.append(XmlUtils.toString(inputDoc.getDocumentElement()));
        }
        if (outputDoc != null) {
            actionInput.append(XmlUtils.toString(outputDoc.getDocumentElement())); 
        }

        EndpointReferenceType memberEPR = entry.getMemberEPR();    
        actionInput.append(
            ObjectSerializer.toString(memberEPR,
                memberEPR.getTypeDesc().getXmlType()));
        actionInput.append(
            ObjectSerializer.toString(config, 
                config.getTypeDesc().getXmlType()));
        
        actionInput.append("</trigger:actionInput>");
        
        if (logger.isDebugEnabled()) {
            logger.debug("Action script input: " + actionInput.toString());
        }
        
        // exec the action script
        Runtime runtime = Runtime.getRuntime();
        Process process = null;
        try {
            process = runtime.exec(commandline);
        } catch(IOException e) {
            throw new Exception("Could not execute " +commandline+" with exception: "+e);
        }        
        
        PrintStream o = new PrintStream(process.getOutputStream());            
        try {
            o.println(actionInput.toString());
        } catch(Exception e) {
            logger.warn("Exception while writing into action script stdin: " + e);
        } finally {
            o.close();
        }

        Document responseDoc = null;
        try {        
            responseDoc = XMLUtils.newDocument(process.getInputStream());
        } catch(Exception e) {
            throw new Exception("Exception while parsing output from action script: " + e);
        }
        
        return responseDoc.getDocumentElement();        
        
    }

    private Document anyToNewDocument(AnyContentType any,
                                      String docRootURI,
                                      String docRootLocalPart) 
        throws Exception
    {
        try 
        {            
            Document responseDoc = XMLUtils.newDocument();
            Node node = responseDoc.appendChild(
                responseDoc.createElementNS(docRootURI, docRootLocalPart));  
            
            Element[] elements = AnyHelper.toElement(any.get_any());
            for(int i=0;elements!=null && i<elements.length;i++) 
            {

                if(elements[i] == null) 
                {
                    if (logger.isDebugEnabled()) {
                        logger.debug("message element "+i+" is null.");
                    }
                } 
                else 
                {
                    node.appendChild(responseDoc.importNode(elements[i], true)); 
                }                           
            }
            return responseDoc;
        }
        catch (Exception e) 
        {
            logger.error("Unhandled exception creating multi-part message document" + e);
            throw e;
        }
    }

    private Document createErrorDocument(String error) 
        throws Exception
    {
        try 
        {            
            Document responseDoc = XMLUtils.newDocument();
            Node node = responseDoc.appendChild(
                responseDoc.createElementNS(
                    TRIGGER_NS,
                    "trigger:actionScriptRuntimeError")
            );
            node.setNodeValue(error);
            return responseDoc;
        }
        catch (Exception e) 
        {
            logger.error("Unhandled exception creating error response document" + e);
            throw e;
        }
    }
    
    private static String getBaseDirectory() 
    {
        return ContainerConfig.getBaseDirectory();
    }

    public void initialize(Object parameters) 
    {
            // NO OP
    }

    public void terminate()
    {
        // NO OP
    }

    public void setSource(AggregatorSource source)
    {
        // NO OP
    }

    public AggregatorSource getSource()
    {
        return null;
    }

}

